#include <stdio.h>
#include "commconnect.h"
#include "lockptr.h"

using namespace std;

TPortException::TPortException(std::string Message)
{
    _Message = Message;
}


TPortException::TPortException(std::string Location, std::string Message)
{
    _Message = Location + string(": ") + Message;
}


TPortException::TPortException(const TPortException &e)
{
    _Message = e._Message;
}


string TPortException::Message(void) const
{
    return _Message;
}


ostream& operator<< (ostream &os, const TPortException &e)
{
    return os << e._Message;
}


TConnection::TConnection()
{
    Active.init (this, &TConnection::get_Active, &TConnection::set_Active);
    BeforeOpen.init(this, &TConnection::get_BeforeOpen, &TConnection::set_BeforeOpen);
    AfterOpen.init(this, &TConnection::get_AfterOpen, &TConnection::set_AfterOpen);
    BeforeClose.init(this, &TConnection::get_BeforeClose, &TConnection::set_BeforeClose);
    AfterClose.init(this, &TConnection::get_AfterClose, &TConnection::set_AfterClose);
    FActive = false;
    FBeforeOpen = false;
    FAfterOpen = false;
    FBeforeClose = false;
    FAfterClose = false;
}


TConnection::~TConnection()
{
    Close();
}


void TConnection::Open()
{
    Active = true;
}


void TConnection::Close()
{
    Active = false;
}


void TConnection::set_Active(bool aEnable)
{
    if (FActive != aEnable)
    {
        if (aEnable)
        {
            DoBeforeOpen();
            try
            {
                if (!OpenConn()) return;
            }

            catch (TPortException &e)
            {
                cerr << "ERROR: " << e << endl;
//throw;
                return;
            }
            FActive = aEnable;
            DoAfterOpen();
        }
        else
        {
            DoBeforeClose();
            CloseConn();
            FActive = aEnable;
            DoAfterClose();
        }
    }
}


bool TConnection::get_Active()
{
    return FActive;

}


void TConnection::CheckActive()
{
    if (!(Active))
        throw TPortException("TConnection::CheckActive()","Inactive Connection !!!");
}


void TConnection::CheckInactive()
{
    if ((Active))
    {
//        cout << "throw" << endl;
        throw TPortException("TConnection::CheckInactive()","Active Connection !!!");
    }
}


TConnectionNotifyEvent TConnection::get_BeforeOpen()
{
    return FBeforeOpen;
}


void TConnection::set_BeforeOpen(TConnectionNotifyEvent value)
{
    FBeforeOpen = value;
}


TConnectionNotifyEvent TConnection::get_AfterOpen()
{
    return FAfterOpen;
}


void TConnection::set_AfterOpen(TConnectionNotifyEvent value)
{
    FAfterOpen = value;

}


TConnectionNotifyEvent TConnection::get_BeforeClose()
{
    return FBeforeClose;
}


void TConnection::set_BeforeClose(TConnectionNotifyEvent value)
{
    FBeforeClose = value;

}


TConnectionNotifyEvent TConnection::get_AfterClose()
{
    return FAfterClose;
}


void TConnection::set_AfterClose(TConnectionNotifyEvent value)
{
    FAfterClose = value;

}


void TConnection::DoBeforeOpen()
{
    if (FBeforeOpen != NULL)
        FBeforeOpen(this);
}


void TConnection::DoAfterOpen()
{
    if (FAfterOpen != NULL)
        FAfterOpen(this);
}


void TConnection::DoBeforeClose()
{
    if (FBeforeClose != NULL)
        FBeforeClose(this);
}


void TConnection::DoAfterClose()
{
    if (FAfterClose != NULL)
        FAfterClose(this);
}


TLogger::TLogger() : TConnection()
{
    FCriticalSection = new pthread_mutex_t;
    pthread_mutex_init(FCriticalSection, NULL);
    AutoOpen.init (this, &TLogger::get_AutoOpen, &TLogger::set_AutoOpen);
    OnException.init (this, &TLogger::get_OnException, &TLogger::set_OnException);
}


TLogger::~TLogger()
{
    pthread_mutex_destroy(FCriticalSection);
    delete FCriticalSection;
}


bool TLogger::get_AutoOpen()
{
    return FAutoOpen;
}


void TLogger::set_AutoOpen(bool aValue)
{
    CheckInactive();
    FAutoOpen = aValue;
}


TOnExceptionEvent TLogger::get_OnException()
{
    return FOnException;
}


void TLogger::set_OnException(TOnExceptionEvent value)
{
    FOnException = value;
}


void TLogger::DoOnException (TPortException& E)
{
    if (FOnException != NULL)
        FOnException(this, E);
}


TCommunicationConnection::TCommunicationConnection(): TLogConnection()
{
    OnRxChar.init (this, &TCommunicationConnection::get_OnRxChar, &TCommunicationConnection::set_OnRxChar);
    FOnRxChar = NULL;
    FDataBind = NULL;
}


TCommRxCharEvent TCommunicationConnection::get_OnRxChar()
{
    return FOnRxChar;
}


void TCommunicationConnection::set_OnRxChar(TCommRxCharEvent value)
{
    FOnRxChar = value;
}


int TCommunicationConnection::Send(std::string S)
{
    unsigned char* tempbuf = (unsigned char*)S.c_str();
    int writechars = Write(tempbuf, S.length());
    return writechars;
}


std::string TCommunicationConnection::Retrieve(int aCount)
{
    unsigned char* tempbuf = (unsigned char*)malloc(aCount+1);
    memset(tempbuf, 0, aCount + 1);
// how many bytes were actually read;
    Read(tempbuf, aCount);
// convert to string
    std::string RetStr = (const char*)tempbuf;
    free(tempbuf);
    return RetStr;

}


void TCommunicationConnection::DoOnRxChar(int Count)
{
    if (FOnRxChar != NULL)
        FOnRxChar(this, Count, FDataBind);
}


TCommHandle::TCommHandle() : TCommunicationConnection()
{
//    hCommDev.init (this, &TCommHandle::get_hCommDev, &TCommHandle::SethCommDev);
    hCommDev.init (this, &TCommHandle::get_hCommDev);
    Baudrate.init (this, &TCommHandle::get_Baudrate, &TCommHandle::set_Baudrate);
    Parity.init (this, &TCommHandle::get_Parity, &TCommHandle::set_Parity);
    Stopbits.init (this, &TCommHandle::get_Stopbits, &TCommHandle::set_Stopbits);
    Databits.init (this, &TCommHandle::get_Databits, &TCommHandle::set_Databits);
    Options.init (this, &TCommHandle::get_Options, &TCommHandle::set_Options);
    FlowControl.init (this, &TCommHandle::get_FlowControl, &TCommHandle::set_FlowControl);

    OnBreak.init (this, &TCommHandle::get_OnBreak, &TCommHandle::set_OnBreak);
    OnCts.init (this, &TCommHandle::get_OnCts, &TCommHandle::set_OnCts);
    OnDsr.init (this, &TCommHandle::get_OnDsr, &TCommHandle::set_OnDsr);
    OnRing.init (this, &TCommHandle::get_OnRing, &TCommHandle::set_OnRing);
    OnRlsd.init (this, &TCommHandle::get_OnRlsd, &TCommHandle::set_OnRlsd);
    OnError.init (this, &TCommHandle::get_OnError, &TCommHandle::set_OnError);
    OnRxFlag.init (this, &TCommHandle::get_OnRxFlag, &TCommHandle::set_OnRxFlag);
    OnTxEmpty.init (this, &TCommHandle::get_OnTxEmpty, &TCommHandle::set_OnTxEmpty);
    OnTxChar.init (this, &TCommHandle::get_OnTxChar, &TCommHandle::set_OnTxChar);

    XonChar.init (this, &TCommHandle::get_XonChar, &TCommHandle::set_XonChar);
    XoffChar.init (this, &TCommHandle::get_XoffChar, &TCommHandle::set_XoffChar);
    ErrorChar.init (this, &TCommHandle::get_ErrorChar, &TCommHandle::set_ErrorChar);
    EofChar.init (this, &TCommHandle::get_EofChar, &TCommHandle::set_EofChar);
    EvtChar.init (this, &TCommHandle::get_EvtChar, &TCommHandle::set_EvtChar);

    ReadTimeout.init (this, &TCommHandle::get_ReadTimeout, &TCommHandle::set_ReadTimeout);
    WriteTimeout.init (this, &TCommHandle::get_WriteTimeout, &TCommHandle::set_WriteTimeout);

    ReadBufSize.init (this, &TCommHandle::get_ReadBufSize, &TCommHandle::set_ReadBufSize);
    WriteBufSize.init (this, &TCommHandle::get_WriteBufSize, &TCommHandle::set_WriteBufSize);
    MonitorEvents.init (this, &TCommHandle::get_MonitorEvents, &TCommHandle::set_MonitorEvents);

    FhCommDev = INVALID_HANDLE_VALUE;
    FReadTimeout = 10000;
    FWriteTimeout = 10000;
    FReadBufSize = 4096;
    FWriteBufSize = 2048;

    TCommEventTypes tempmonitorevents;
    tempmonitorevents.insert(evBreak);
    tempmonitorevents.insert(evCts);
    tempmonitorevents.insert(evDsr);
    tempmonitorevents.insert(evError);
    tempmonitorevents.insert(evRing);
    tempmonitorevents.insert(evRlsd);
    tempmonitorevents.insert(evRxChar);
    tempmonitorevents.insert(evRxFlag);
    tempmonitorevents.insert(evTxEmpty);
    tempmonitorevents.insert(evTxChar);
    FMonitorEvents = tempmonitorevents;

    FBaudrate = br9600;

    FParity = paNone;

    FStopbits = sb10;

    FDatabits = da8;

    FOptions.erase(coParityCheck);

    FOptions.erase(coDsrSensitivity);
    FOptions.erase(coIgnoreXOff);
    FOptions.erase(coErrorChar);
    FOptions.erase(coNullStrip);

    FFlowControl = fcDefault;

    FXonChar = (char)17;

    FXoffChar = (char)19;

    FOnBreak = NULL;
    FOnCts = NULL;
    FOnDsr = NULL;
    FOnRing = NULL;
    FOnRlsd = NULL;
    FOnRxFlag = NULL;
    FOnTxEmpty = NULL;
    FOnTxChar = NULL;
    FOnError = NULL;

    FCriticalSection = new pthread_mutex_t;

    pthread_mutex_init(FCriticalSection, NULL);

}


TCommHandle::~TCommHandle()
{
    pthread_mutex_destroy(FCriticalSection);
    delete FCriticalSection;
}


TComHandle TCommHandle::get_hCommDev()
{
    return FhCommDev;
}


void TCommHandle::SethCommDev(TComHandle value)
{
    CheckInactive();
    FhCommDev = value;
}


TBaudrate TCommHandle::get_Baudrate()
{
    return FBaudrate;
}


void TCommHandle::set_Baudrate(TBaudrate value)
{
    FBaudrate = value;
    UpdateDataControlBlock();
}


TParity TCommHandle::get_Parity()
{
    return FParity;
}


void TCommHandle::set_Parity(TParity value)
{
    FParity = value;
    UpdateDataControlBlock();
}


TStopbits TCommHandle::get_Stopbits()
{
    return FStopbits;
}


void TCommHandle::set_Stopbits(TStopbits value)
{
    FStopbits = value;
    UpdateDataControlBlock();
}


TDatabits TCommHandle::get_Databits()
{
    return FDatabits;
}


void TCommHandle::set_Databits(TDatabits value)
{
    FDatabits = value;
    UpdateDataControlBlock();
}


TCommOptions TCommHandle::get_Options()
{
    return FOptions;
}


void TCommHandle::set_Options(TCommOptions value)
{
    FOptions = value;
    UpdateDataControlBlock();
}


TFlowControl TCommHandle::get_FlowControl()
{
    return FFlowControl;
}


void TCommHandle::set_FlowControl(TFlowControl value)
{
    FFlowControl = value;
    UpdateDataControlBlock();
}


TNotifyEvent TCommHandle::get_OnBreak()
{
    return FOnBreak;
}


void TCommHandle::set_OnBreak(TNotifyEvent value)
{
    FOnBreak = value;
}


TNotifyEvent TCommHandle::get_OnCts()
{
    return FOnCts;
}


void TCommHandle::set_OnCts(TNotifyEvent value)
{
    FOnCts = value;
}


TNotifyEvent TCommHandle::get_OnDsr()
{
    return FOnDsr;
}


void TCommHandle::set_OnDsr(TNotifyEvent value)
{
    FOnDsr = value;
}


TNotifyEvent TCommHandle::get_OnRing()
{
    return FOnRing;
}


void TCommHandle::set_OnRing(TNotifyEvent value)
{
    FOnRing = value;
}


TNotifyEvent TCommHandle::get_OnRlsd()
{
    return FOnRlsd;
}


void TCommHandle::set_OnRlsd(TNotifyEvent value)
{
    FOnRlsd = value;
}


TCommErrorEvent TCommHandle::get_OnError()
{
    return FOnError;
}


void TCommHandle::set_OnError(TCommErrorEvent value)
{
    FOnError = value;
}


TNotifyEvent TCommHandle::get_OnRxFlag()
{
    return FOnRxFlag;
}


void TCommHandle::set_OnRxFlag(TNotifyEvent value)
{
    FOnRxFlag = value;
}


TNotifyEvent TCommHandle::get_OnTxEmpty()
{
    return FOnRxFlag;
}


void TCommHandle::set_OnTxEmpty(TNotifyEvent value)
{
    FOnTxEmpty = value;
}


TNotifyEvent TCommHandle::get_OnTxChar()
{
    return FOnTxChar;
}


void TCommHandle::set_OnTxChar(TNotifyEvent value)
{
    FOnTxChar = value;
}


char TCommHandle::get_XonChar ()
{
    return FXonChar;
}


void TCommHandle::set_XonChar (char value)
{
    FXonChar = value;
    UpdateDataControlBlock();
}


char TCommHandle::get_XoffChar ()
{
    return FXoffChar;
}


void TCommHandle::set_XoffChar (char value)
{
    FXoffChar = value;
    UpdateDataControlBlock();
}


char TCommHandle::get_ErrorChar ()
{
    return FErrorChar;
}


void TCommHandle::set_ErrorChar (char value)
{
    FErrorChar = value;
    UpdateDataControlBlock();
}


char TCommHandle::get_EofChar ()
{
    return FEofChar;
}


void TCommHandle::set_EofChar (char value)
{
    FEofChar = value;
    UpdateDataControlBlock();
}


char TCommHandle::get_EvtChar ()
{
    return FEvtChar;
}


void TCommHandle::set_EvtChar (char value)
{
    FEvtChar = value;
    UpdateDataControlBlock();
}


int TCommHandle::get_ReadTimeout ()
{
    return FReadTimeout;
}


void TCommHandle::set_ReadTimeout (int value)
{
    FReadTimeout = value;
}


int TCommHandle::get_WriteTimeout ()
{
    return FWriteTimeout;
}


void TCommHandle::set_WriteTimeout (int value)
{
    FWriteTimeout = value;
}


int TCommHandle::get_ReadBufSize ()
{
    return FReadBufSize;
}


void TCommHandle::set_ReadBufSize (int value)
{
    CheckInactive();
    FReadBufSize = value;
}


int TCommHandle::get_WriteBufSize ()
{
    return FWriteBufSize;
}


void TCommHandle::set_WriteBufSize (int value)
{
    CheckInactive();
    FWriteBufSize = value;
}


TCommEventTypes TCommHandle::get_MonitorEvents()
{
    return FMonitorEvents;
}


void TCommHandle::set_MonitorEvents(TCommEventTypes value)
{
    CheckInactive();
    FMonitorEvents = value;
}


void TCommHandle::UpdateDataControlBlock()
{
    if (Active)
        UpdateDCB();
}


bool TCommHandle::IsErrorMonitored ()
{
    TCommEventTypes::iterator ScopeIterator;
    for (ScopeIterator = FMonitorEvents.begin(); ScopeIterator != FMonitorEvents.end(); ScopeIterator++)
    {
        if (*ScopeIterator == evError)
            return true;
    }
    return false;
}


void TCommHandle::UpdateDCB()
{
    struct termios Term;

    const int CommBaudRates[15] =
    {
        B110, B300, B600, B1200, B2400, B4800, B9600, -1,
        B19200,
#ifdef B38400
        B38400,
#else
        -1,
#endif
        -1,
#ifdef B57600
        B57600,
#else
        -1,
#endif
#ifdef B115200
        B115200,
#else
        -1,
#endif
        -1,
#ifdef B230400
        B230400
#else
        -1
#endif
    };

    const int CommDataBits[5] = {-1, CS5, CS6, CS7, CS8};
    tcgetattr(FhCommDev, &Term);
#ifndef __CYGWIN__
    cfmakeraw (&Term);
#else
    Term.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP |INLCR|IGNCR|ICRNL|IXON);
    Term.c_oflag &= ~OPOST;
    Term.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
    Term.c_cflag &= ~(CSIZE|PARENB);
    Term.c_cflag |= CS8;
#endif
    Term.c_iflag = Term.c_iflag | IGNBRK;
    if (IsErrorMonitored())
        Term.c_iflag = Term.c_iflag & (~IGNPAR) | PARMRK;
    else
        Term.c_iflag = Term.c_iflag | IGNPAR;

    if ((FParity == paOdd) || (FParity == paEven))
        Term.c_iflag = Term.c_iflag | INPCK;
    else
        Term.c_iflag = Term.c_iflag & (~INPCK);

    if (FFlowControl == fcSoftware)
        Term.c_iflag = Term.c_iflag | (IXON | IXOFF | IXANY);
    else if (FFlowControl != fcDefault)
        Term.c_iflag = Term.c_iflag & (~(IXON | IXOFF | IXANY));

// control flags
    Term.c_cflag = Term.c_cflag | CREAD | HUPCL | CLOCAL;

    Term.c_cflag = Term.c_cflag & (~CSIZE);
    if (CommDataBits[FDatabits] == -1)
        throw TPortException("TCommHandle::UpdateDCB()","CommErrDatabits");

    Term.c_cflag = Term.c_cflag & (~CSIZE) | CommDataBits[FDatabits];

    if (FStopbits == sb10) Term.c_cflag = Term.c_cflag & (~CSTOPB); else
        if (FStopbits == sb15) throw TPortException("TCommHandle::UpdateDCB()","CommErrStopbits"); else
            Term.c_cflag = Term.c_cflag | CSTOPB;

    if (FParity == paNone)
        Term.c_cflag = Term.c_cflag & (~PARENB);
    else
        Term.c_cflag = Term.c_cflag | PARENB;

    switch (FParity)
    {
        case paOdd:
            Term.c_cflag = Term.c_cflag | PARODD;
        case paEven:
            Term.c_cflag = Term.c_cflag & (~PARODD);
        case paMark:
            throw TPortException("TCommHandle::UpdateDCB()","CommErrParity");
        case paSpace:
            throw TPortException("TCommHandle::UpdateDCB()","CommErrParity");
        default: {break;}
    }

    if (FFlowControl == fcCTS)
        Term.c_cflag = Term.c_cflag | CRTSCTS;
    else if (FFlowControl == fcDTR) throw TPortException("TCommHandle::UpdateDCB()","CommErrFlow");
    else if (FFlowControl != fcDefault)
        Term.c_cflag = Term.c_cflag & (~CRTSCTS);

    if (CommBaudRates[FBaudrate] == -1)
        throw TPortException("TCommHandle::UpdateDCB()","CommErrBaudrate");

    cfsetospeed (&Term, CommBaudRates[FBaudrate]);
    cfsetispeed (&Term, CommBaudRates[FBaudrate]);

// local modec
    Term.c_lflag = Term.c_lflag & (~ICANON);

// character slots
    Term.c_cc[VEOF] = EofChar;
    Term.c_cc[VSTART] = XonChar;
    Term.c_cc[VSTOP] = XoffChar;
    Term.c_cc[VINTR] = EvtChar;

    Term.c_cc[VMIN] = (char)0;
    Term.c_cc[VTIME] = (char)0;

#ifndef __QNX__
    if (tcsetattr (FhCommDev, TCSANOW, &Term))
#else
        if (tcsetattr (FhCommDev, TCSAFLUSH, &Term))
#endif
            throw TPortException("TCommHandle::UpdateDCB()","tcsetattr TCSANOW");
}


bool TCommHandle::GetModemState (int index)
{
    bool result = false;
    unsigned long Flag;
    if (Active)
    {
        if (ioctl (FhCommDev, TIOCMGET, &Flag) < 0)
            throw TPortException("TCommHandle::GetModemState()","ioctl TIOCMGET");
        result = ((Flag & index) != 0);
    }

    return result;
}


void TCommHandle::SetEscBreak(bool value)
{
    if (value)
        tcsendbreak(FhCommDev, 0);
}


void TCommHandle::SetEsc(int Index, bool value)
{
    if (Active)
    {
        unsigned long Flags;
        const unsigned long Esc[3] = {TIOCM_DTR, TIOCM_RTS, 0};
        if (ioctl(FhCommDev, TIOCMGET, &Flags) == 0)
        {
            if (value)
                Flags = Flags | Esc[Index-1];
            else
                Flags = Flags & (~Esc[Index-1]);
            ioctl(FhCommDev, TIOCMSET, &Flags);
            if (Active && (Index == 3))
                ioctl(FhCommDev, TCFLSH, TCIOFLUSH);
        }
    }
}


void TCommHandle::SetDTRState(bool value)
{
    SetEsc(1, value);
}


void TCommHandle::SetRTSState(bool value)
{
    SetEsc(2, value);
}


void TCommHandle::Bind(void* DataBind)
{
//    if (FDataBind == NULL)
    FDataBind = DataBind;
}


void TCommHandle::UpdateCommTimeouts()
{

}


typedef struct TCommonStruct
{
    unsigned long EventThread;
    unsigned long CommHandle;
};

void* run (void* arg);

bool TCommHandle::OpenConn()
{
    if (FhCommDev == INVALID_HANDLE_VALUE)
    {
        cerr << "ERROR: TCommHandle::OpenConn() - CreateFile" << endl;
        return false;
    }

    UpdateCommTimeouts();
    UpdateDCB();

    FEventThread = new TCommEventThread (this, FhCommDev, FMonitorEvents);

#ifdef NO_THREAD_ADAPTER
    unsigned long* pDataLink = new (unsigned long);
    *pDataLink = (unsigned long) FEventThread;
#endif

#ifdef __CYGWIN__
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
#ifndef NO_THREAD_ADAPTER
    pthread_create(&t, &attr, ADAPTER(TCommEventThread,run), FEventThread);
#else
    pthread_create(&t, &attr, &run, pDataLink);
#endif
    pthread_attr_destroy(&attr);
#else
#ifndef NO_THREAD_ADAPTER
    pthread_create(&t, NULL, ADAPTER(TCommEventThread, run), FEventThread);
#else
    pthread_create(&t, NULL, &run, pDataLink);
#endif
#endif
//    UpdateCommTimeouts();
//    UpdateDCB();
    return true;
}


void TCommHandle::CloseConn()
{
#ifdef COMMDEBUG
    cout << "TCommHandle::CloseConn()" << endl;
#endif
    if (FhCommDev != INVALID_HANDLE_VALUE)
    {
        FEventThread->Terminate();

#ifndef __CYGWIN__
        pthread_join(t, NULL);
#else
        do {sched_yield();}
        while (!FEventThread->ConsiderTerminated());
#endif

        ::close(FhCommDev);
        FhCommDev = INVALID_HANDLE_VALUE;
    }
}


void TCommHandle::Lock()
{
    pthread_mutex_lock(FCriticalSection);
}


void TCommHandle::Unlock()
{
    pthread_mutex_unlock(FCriticalSection);
}


unsigned long TCommHandle::GetTickCount(void)
{
    unsigned long ticks;

    struct timeval now;
    gettimeofday(&now, NULL);
    ticks = now.tv_sec * 1000l;
    ticks += now.tv_usec / 1000l;
    return ticks;
}


int TCommHandle::Write (unsigned char* Buf, int Count)
{
    unsigned long Tick;
    unsigned char* P;
    int Result;

    if (!Active)
        return 0;

    Lock();

    FEventThread->WriteFlag = true;
    Tick = GetTickCount();
    P = Buf;
    do
    {
        Result = ::write (FhCommDev, P, Count);
#ifdef __QNX__
//        if (Result > 1)
//          tcdrain (FhCommDev);
//	cout << Result << endl;
//	cout << "wr3" << endl;
#endif
        if (Result > 0)
        {
            P += Result;
            Count -= Result;
        }
    } while ((Result >= 0) && (Count > 0) && (FWriteTimeout != 0) && (abs( GetTickCount()-Tick < (unsigned long)FWriteTimeout )) );
    if (TComHandle(Result) == INVALID_HANDLE_VALUE)
    {
        cout << "TCommHandle::Write() - FileWrite" << endl;
        Unlock();
//        throw TPortException("TCommHandle::Write()","FileWrite");  // ????
    }

    Unlock();
    return Result;
}


int TCommHandle::Read (unsigned char* Buf, int Count)
{
    unsigned long Tick;
    unsigned char* P;
    int Result;

    if (!Active)
        return 0;

    Lock();

//FEventThread->WriteFlag = false;
    Tick = GetTickCount();
    P = Buf;
    do
    {
        Result = ::read (FhCommDev, P, Count);
        if (Result > 0)
        {
            P += Result;
            Count -= Result;
        }
    } while ((Result >= 0) && (Count > 0) && (FReadTimeout != 0) && (int)abs((GetTickCount()-Tick < (unsigned long)FReadTimeout)));

    if (TComHandle(Result) == INVALID_HANDLE_VALUE)
    {
        cout << "TCommHandle::Read() - FileRead" << endl;
        Unlock();
//???        throw TPortException("TCommHandle::Read()","FileRead");
    }

    Unlock();
    return Result;
}


#ifndef TIOCINQ
#define TIOCINQ FIONREAD
#endif

int TCommHandle::InQueCount()
{
    int Result = 0;
    if (Active)
    {
//        Lock();
        ::ioctl(int(FhCommDev), TIOCINQ, &Result);
//        Unlock();
    };

    return Result;
}


int TCommHandle::OutQueCount()
{
    int Result = 0;
    if (Active)
    {
//        Lock();
#ifndef __CYGWIN__
        ::ioctl(int(FhCommDev), TIOCOUTQ, &Result);
#else
        ::ioctl(int(FhCommDev), 0x5411, &Result);
#endif
//        Unlock();
    }

    return Result;
}


void TCommHandle::PurgeIn()
{
    if (Active)
    {
//        ::ioctl(int(FhCommDev), TCFLSH, TCIFLUSH);
        Lock();
        tcflush(FhCommDev, TCIFLUSH);
        Unlock();
    }
}


void TCommHandle::PurgeOut()
{
    if (Active)
    {
//        cout << "OutQueCount " << OutQueCount() << endl;
//        ::ioctl(int(FhCommDev), TCFLSH, TCOFLUSH);
        Lock();
        tcflush(FhCommDev, TCOFLUSH);
        Unlock();
//	sleep(1);
//	cout << "OutQueCount " << OutQueCount() << endl;
    }
}


const unsigned long EV_RXCHAR = 1;
const unsigned long EV_RXFLAG = 2;
const unsigned long EV_TXEMPTY = 4;
const unsigned long EV_CTS = 8;
const unsigned long EV_DSR = 0x10;
const unsigned long EV_RLSD = 0x20;
const unsigned long EV_BREAK = 0x40;
const unsigned long EV_ERR = 0x80;
const unsigned long EV_RING = 0x100;
const unsigned long EV_PERR = 0x200;
const unsigned long EV_TXCHAR = 0x400;

void HandleCommEvent(unsigned long status, TCommHandle* commhandle)
{

    unsigned long Errors = 0;
    int N;

    if ((status & EV_BREAK) > 0)
        if (commhandle->FOnBreak)
            commhandle->FOnBreak(commhandle);

    if ((status & EV_CTS) > 0)
        if (commhandle->FOnCts)
            commhandle->FOnCts(commhandle);

    if ((status & EV_DSR) > 0)
        if (commhandle->FOnDsr)
            commhandle->FOnDsr(commhandle);

    if ((status & EV_ERR) > 0)
        if (commhandle->FOnError)
            commhandle->FOnError(commhandle, Errors);

    if ((status & EV_RING) > 0)
        if (commhandle->FOnRing)
            commhandle->FOnRing(commhandle);

    if ((status & EV_RLSD) > 0)
        if (commhandle->FOnRlsd)
            commhandle->FOnRlsd(commhandle);

    if ((status & EV_RXCHAR) > 0)
    {
        ::ioctl((int)commhandle->FhCommDev, TIOCINQ, &N);
        if (N > 0) commhandle->DoOnRxChar(N);
    }

    if ((status & EV_RXFLAG) > 0)
        if (commhandle->FOnRxFlag)
            commhandle->FOnRxFlag(commhandle);

    if ((status & EV_TXEMPTY) > 0)
        if (commhandle->FOnTxEmpty)
            commhandle->FOnTxEmpty(commhandle);

    if ((status & EV_TXCHAR) > 0)
        if (commhandle->FOnTxChar)
            commhandle->FOnTxChar(commhandle);

    if ((!commhandle->FOnBreak) || (!commhandle->FOnCts) ||
        (!commhandle->FOnDsr) || (!commhandle->FOnError) ||
        (!commhandle->FOnRing) || (!commhandle->FOnRlsd) ||
        (!commhandle->FOnRxFlag) || (!commhandle->FOnTxEmpty))
        usleep(1);
}


TComm::TComm() : TCommHandle()
{
//cout << "TComm"<< endl;
    DeviceName.init (this, &TComm::get_DeviceName, &TComm::set_DeviceName);
    FDeviceName = DefaultDeviceName;
}


TComm::~TComm()
{
    TConnection::Close();
#ifdef COMMDEBUG
    cout << "TComm::~TComm()" << endl;
#endif
}


std::string TComm::get_DeviceName()
{
    return FDeviceName;
}


void TComm::set_DeviceName(std::string value)
{
    CheckInactive();
    FDeviceName = value;
}


bool TComm::OpenConn()
{
    FhCommDev = open((const char*)FDeviceName.c_str(), O_RDWR | O_NOCTTY | O_NONBLOCK);
// invalid handle (detected further)
    if (FhCommDev == INVALID_HANDLE_VALUE) return false;
    return TCommHandle::OpenConn();
}


void TComm::CloseConn()
{
#ifdef COMMDEBUG
    cout << "TComm::CloseConn()" << endl;
#endif
    if (FhCommDev != INVALID_HANDLE_VALUE)
        TCommHandle::CloseConn();
/// ???
    TCommHandle::CloseConn();
}


void TCommEventThread::DoOnSignal()
{
    unsigned long Msk;
    pthread_mutex_lock(FCriticalSection);
    Msk = EventMask;
    EventMask = 0;
    pthread_mutex_unlock(FCriticalSection);
    HandleCommEvent(Msk, FComm);
}


TCommEventThread::TCommEventThread(TCommHandle* aComm, TComHandle Handle, TCommEventTypes Events)
{
    ioctl(FCommHandle, TIOCMGET, &FModemFlags);
    FEvents = Events;
    CommHandle.init (this, &TCommEventThread::get_CommHandle);
    Terminated.init (this, &TCommEventThread::get_Terminated);
    CriticalSection.init (this, &TCommEventThread::get_CriticalSection);
    Comm.init (this, &TCommEventThread::get_Comm);
    FCommHandle = Handle;
    FTerminated = false;
    FCriticalSection = new pthread_mutex_t;
    FComm = aComm;
    pthread_mutex_init(FCriticalSection, NULL);

#ifdef __CYGWIN__
    pthread_mutex_init (&FConsiderTerminatedMutex, NULL);
#endif
}


TCommEventThread::~TCommEventThread()
{
#ifdef __CYGWIN__
    pthread_mutex_destroy (&FConsiderTerminatedMutex);
#endif

    pthread_mutex_destroy(FCriticalSection);
    delete FCriticalSection;
}


void TCommEventThread::Terminate()
{
    FTerminated = true;
}


TComHandle TCommEventThread::get_CommHandle()
{
    return FCommHandle;
}


bool TCommEventThread::get_Terminated()
{
    return FTerminated;
}


pthread_mutex_t* TCommEventThread::get_CriticalSection()
{
    return FCriticalSection;
}


TCommHandle* TCommEventThread::get_Comm()
{
    return FComm;
}


#ifdef __CYGWIN__
bool TCommEventThread::ConsiderTerminated()
{
    LockingPtr<bool> lpConsiderTerminated(FConsiderTerminated, FConsiderTerminatedMutex);
    return  *lpConsiderTerminated;
}
#endif

void* TCommEventThread::run()
{
    fd_set rfds, wfds, efds;
    fd_set* wfds2;
    struct timeval tv;
    int selres;
//cout << "1" << endl;
#ifdef __CYGWIN__
    LockingPtr<bool> lpConsiderTerminated(FConsiderTerminated, FConsiderTerminatedMutex);
    *lpConsiderTerminated = false;
#endif
    while (!FTerminated)
    {
        FD_ZERO(&rfds); FD_ZERO(&wfds); FD_ZERO(&efds);
        FD_SET(FCommHandle, &rfds);
        FD_SET(FCommHandle, &wfds);
        FD_SET(FCommHandle, &efds);
// was 0 ???
        tv.tv_sec = 1;
// was 10 ???
        tv.tv_usec = 0;
        if (WriteFlag)
            wfds2 = &wfds;
        else
            wfds2 = NULL;

        if ((selres = select (FCommHandle+1, &rfds, wfds2, &efds, &tv)) > 0)
        {
// > 0 have input
// = 0 timeout
// < 0 select failed

            pthread_mutex_lock(FCriticalSection);

            if (FD_ISSET(FCommHandle, &rfds))
                EventMask = EventMask | EV_RXCHAR;
            if ((WriteFlag) && (FD_ISSET(FCommHandle, &wfds)))
            {
                if (FComm->OutQueCount() == 0)
                {
                    EventMask = EventMask | EV_TXEMPTY;
                    WriteFlag = false;
                }
                else
                {
                    EventMask = EventMask | EV_TXCHAR;
                }
            }
            else
            {

            }
            if (FD_ISSET(FCommHandle, &efds))
                EventMask = EventMask | EV_ERR;

            pthread_mutex_unlock(FCriticalSection);

            if (EventMask != 0)
            {
                DoOnSignal();
            }
            else
            {
#ifdef __CYGWIN__
//                EventMask = EventMask | EV_RXCHAR;
#endif
            }
        }
        else
        {
/* if (selres == 0) cout << "t_out" << endl; */
        }

    }
    ioctl (FCommHandle, TCFLSH, TCIOFLUSH);
#ifdef __CYGWIN__
    *lpConsiderTerminated = true;
#endif

    return NULL;
}


void* run(void* arg)
{
    fd_set rfds, wfds, efds;
    fd_set* wfds2;
    struct timeval tv;
    int selres;

//cout << "2" << endl;

    TCommEventThread* CommEventThread = *((TCommEventThread**)arg);

#ifdef __CYGWIN__
    LockingPtr<bool> lpConsiderTerminated(CommEventThread->FConsiderTerminated, CommEventThread->FConsiderTerminatedMutex);
    *lpConsiderTerminated = false;
#endif
    while (!CommEventThread->Terminated)
    {
        FD_ZERO(&rfds); FD_ZERO(&wfds); FD_ZERO(&efds);
        FD_SET(CommEventThread->CommHandle, &rfds);
        FD_SET(CommEventThread->CommHandle, &wfds);
        FD_SET(CommEventThread->CommHandle, &efds);
// was 0 ???
        tv.tv_sec = 1;
// was 10 ???
        tv.tv_usec = 0;
        if (CommEventThread->WriteFlag)
            wfds2 = &wfds;
        else
            wfds2 = NULL;
        if ((selres = select (CommEventThread->CommHandle+1, &rfds, wfds2, &efds, &tv)) > 0)
        {
// > 0 have input
// = 0 timeout
// < 0 select failed

            pthread_mutex_lock(CommEventThread->CriticalSection);

            if (FD_ISSET(CommEventThread->CommHandle, &rfds))
                CommEventThread->EventMask = CommEventThread->EventMask | EV_RXCHAR;
            if ((CommEventThread->WriteFlag) && (FD_ISSET(CommEventThread->CommHandle, &wfds)))
            {
                if (((TCommHandle*)(CommEventThread->Comm))->OutQueCount() == 0)
                {
                    CommEventThread->EventMask = CommEventThread->EventMask | EV_TXEMPTY;
                    CommEventThread->WriteFlag = false;
                }
                else
                {
                    CommEventThread->EventMask = CommEventThread->EventMask | EV_TXCHAR;
                }
            }
            else
            {
// ???  usleep(1);
            }
            if (FD_ISSET(CommEventThread->CommHandle, &efds))
                CommEventThread->EventMask = CommEventThread->EventMask | EV_ERR;

            pthread_mutex_unlock(CommEventThread->CriticalSection);

            if (CommEventThread->EventMask != 0)
            {
                CommEventThread->DoOnSignal();
            }
            else
            {
#ifdef __CYGWIN__
//                CommEventThread->EventMask = CommEventThread->EventMask | EV_RXCHAR;
#endif
            }
        }
        else
        {
/* if (selres == 0) cout << "t_out" << endl; */
        }

    }
    ioctl (CommEventThread->CommHandle, TCFLSH, TCIOFLUSH);

#ifdef __CYGWIN__
    *lpConsiderTerminated = true;
#endif
    TCommEventThread** ppDataLink = (TCommEventThread**)arg;
    delete ppDataLink;

    return NULL;
}


///
