#ifndef __LOCKPTR_H
#define __LOCKPTR_H


#ifdef __CYGWIN__
template <typename T>
class LockingPtr
{
    public:
        LockingPtr(volatile T& obj, pthread_mutex_t& mtx)
            : pObj_(const_cast<T*>(&obj)),
            pMtx_(&mtx)
            { pthread_mutex_lock(&mtx); }
        ~LockingPtr()
            { pthread_mutex_unlock(pMtx_); }
// Pointer behaviour
        T& operator*()
            { return *pObj_; }
        T* operator->()
            { return pObj_; }
    private:
        T* pObj_;
        pthread_mutex_t* pMtx_;
        LockingPtr(const LockingPtr&);
        LockingPtr& operator= (const LockingPtr&);
};
#endif

#endif
